-- Copyright (c) 2012 Takahiro Shinozaki
-- 
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
-- 
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
-- 
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-- THE SOFTWARE.

--------------------------------------------------------------------------------
-- |
-- Module      :  DcdComps
-- Copyright   :  Takahiro Shinozaki 2012
-- License     :  MIT
-- 
-- Maintainer  :  www.ailab.tj.chiba-u.jp/~shinot
-- Stability   :  experimental
-- Portability :  portable
-- 
-- This module contains definitions of pipeline components for speech decoding.
-- This is the mainbody of the husky decoder.
--------------------------------------------------------------------------------


module DcdComps (
  -- * Decoding configuration
  DcdConfig(insPenalty, lmWeight, band, beam), 
  readConfig,
  configPsr,
  -- * Main decoding functions
  decodeForward,
  progressiveBacktrack,
  -- * Sub functions
  wfstExpandFrame,
  wfstExpandEps,
  pruneHypotheses
  )
       where

import Debug.Trace

import qualified Data.ByteString.Lazy.Char8 as L8
import qualified Data.List as List
import qualified Data.Vector as V
-- import Data.Function.Memoize
import Data.Binary.Get
import WFST
import GMM
import MHUtils

import Control.Monad.ST
import qualified Data.Vector.Mutable as MV
import qualified Data.Vector.Generic as VG
import qualified Data.Vector.Algorithms.Intro as VA

-- Config params ----------------------------------------
-- | A holder of a set of decoding parameters
data DcdConfig = DcdConfig {
  insPenalty :: Double, -- insertion penalty  
  lmWeight :: Double, -- language model weight
  band :: Int, -- band width for pruning (number of hypothesis)
  beam :: Double -- beam width for pruning (score difference from the top score)
  } deriving (Show, Eq)
             
-- | Takes a file name of a configuration file for decoding and reads it.
-- Each line of the file specifies a name-value pair: name=val.
-- The parameters include insPenalty, lmWeight, band, and beam.
readConfig :: String -- ^ A name of configuration file 
              -> IO DcdConfig -- ^ configuration
readConfig fname = do {buf <- L8.readFile fname; 
                       return (List.foldl configPsr (DcdConfig 0 10 10000 200) (map L8.unpack (L8.words buf)))}

-- | Parse a string describing parameters. 
-- Returns a DcdConfig where one of the data fields is updated following the specification in the string.
configPsr :: DcdConfig -- ^ Default value
             -> String -- ^ String to parse
             -> DcdConfig -- ^ Updated value
configPsr (DcdConfig insp lmw bnd bm) definition = (DcdConfig mInsp mLmw mBnd mBm)
-- This is not efficient as all the variables are checked every time. Assumes the config file is small.
  where mInsp = getVal "insPenalty=" insp
        mLmw = getVal "lmWeight=" lmw
        mBnd = getVal "band=" bnd
        mBm = getVal "beam=" bm
        getVal defname dfv = case List.stripPrefix defname definition of 
          Just val -> read val
          _ -> dfv

-- Pipeline components for decoding  --------------------------------------
-- These takes one or more lists and returns a list

-- | Performs Viterbi forward pass for speech decoding.
decodeForward :: DcdConfig -- ^ Decoding configuration.
                 -> WFST -- ^ WFST definitions.
                 -> V.Vector GMixture -- ^ A set of Gaussian mixture definitions of HMM states.
                 -> [FeaVector] -- ^ A time series of a feature vector.
                 -> PathList -- ^ Final set of paths.
decodeForward config wfst spdfs feaSeq = foldl step ini feaSeq
  where ini = pruneHypotheses config $ expandPathListForEps wfst $ 
              pathListFromList $ [pathFromList [(initArc (iniSt wfst) (iniSt wfst) (-1) (-1) 0.0)]] 

        step :: PathList -> FeaVector -> PathList
        step ps fea = pruneHypotheses config $ wfstExpandEps wfst $ wfstExpandFrame wfst spdfs fea ps
        
-- | Performs progressive backtracking for every specified interval frames. 
-- Determined arcs are output progressively without waiting for the end of the input.
-- Large interval results in coarse output timing and small one results in large 
-- computational overhed for frequent partial backtracking.
progressiveBacktrack :: Int -- ^ Interval to perform partial backtracking. Specified in frames.
                        -> Int -- ^ State ID of the final WFST state.
                        -> PathList -- ^ A set of paths
                        -> [Arc] -- ^ A time series of an arc that has been determined. i.e. recognition results.
progressiveBacktrack intval finSt lat = arcSeq
  where arcSeq = pathToList fpath
        fpath = head $ filter (\p -> pathEd p == finSt) lat

-- | Expands candidate states at each time frame. Epsilon transitions are not expanded.
wfstExpandFrame :: WFST -- ^ WFST definitions
                   -> V.Vector GMixture  -- ^ A set of Gaussian mixture definitions of HMM states.
                   -> FeaVector -- ^ A feature vector.
                   -> PathList -- ^ A set of candidate paths so far with their accumulated scores.
                   -> PathList -- ^ A time series of a set of frame hypotheses obtained by expanding the candidate states.
wfstExpandFrame wfst spdfs fea pl = expandPathListWithILblScore wfst pl mHmmObsScr
  where
        mHmmObsScr = memoize $ hmmObsScr fea
        hmmObsScr :: FeaVector -> Int -> Double
        hmmObsScr xf st = -(gmmLProb (spdfs V.! (st-1)) xf)

-- | Expands a set of frame hypotheses by epsilon transitions.
wfstExpandEps :: WFST -- ^ WFST definitions
                 -> PathList -- ^ A set of frame hypotheses.
                 -> PathList -- ^ A list of (possibly extended) candidate paths with accumulated scores, sorted by score
wfstExpandEps wfst pl = expandPathListForEps wfst pl

-- | Prune hypotheses
pruneHypotheses :: DcdConfig -- ^ Decoding configuration.
                   -> PathList -- ^ A list of candidate paths, unsorted.
                   -> PathList -- ^ Pruned results, sorted by score.
pruneHypotheses config pl = prunePL $ sortOut pl
  where prunePL x = pathListTakeWhileScore (\scr -> (topScr x) + (beam config) > scr) $ -- Beam pruning
                    pathListTake (band config) $ -- Band pruning
                    x
        topScr x  = case pathListSize x of 
                     0 -> 0.0
                     _ -> pathScore $ List.head $ pathListToList $ pathListTake 1 x

        sortOut :: PathList -> PathList
        sortOut ps = runST $ do
           -- A piece of FORTRAN in Haskell
           mv <- MV.new (length ps)
           mapM_ (\ (i, p) -> MV.write mv i p) (zip [0..] ps)
           doSort mv
           v <- VG.unsafeFreeze mv
           return $ V.toList v

        -- A bug in partialSortBy: it does not check that the vector is long enough.
        doSort mv = VA.partialSortBy pathCompare mv (min (band config) (MV.length mv))

        pathCompare :: Path -> Path -> Ordering
        pathCompare p1 p2 = compare (pathScore p1) (pathScore p2)

