-- Copyright (c) 2012 Takahiro Shinozaki
-- 
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
-- 
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
-- 
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-- THE SOFTWARE.

--------------------------------------------------------------------------------
-- |
-- Module      :  GMM
-- Copyright   :  Takahiro Shinozaki 2012
-- License     :  MIT
-- 
-- Maintainer  :  www.ailab.tj.chiba-u.jp/~shinot
-- Stability   :  experimental
-- Portability :  portable
-- 
-- A Gaussian mixture (GMM) module. 
--------------------------------------------------------------------------------

{-# LANGUAGE ForeignFunctionInterface #-}
{-# LANGUAGE BangPatterns #-}

-- C header file 
#include "gmmclib.h"

module GMM (
  GMixture,
  FeaVector,
  HtkFea,
  spdfsFromFile,
  inspectSPDFs,
  readSpdfs,
  htkParmFromFile,
  txtFeaFromFile,
  feaVec,
  gmmLProb,
  ) where

import Foreign.C.Types -- (CFloat, CDouble, CInt, CLong) -- restriction fails for some reason -- goga
import Foreign.Ptr (Ptr, FunPtr, freeHaskellFunPtr)
import System.IO.Unsafe (unsafePerformIO)
-- import System
import qualified Data.ByteString.Lazy.Char8 as L8
import Data.ByteString.Lex.Lazy.Double
import qualified Data.Vector.Storable as SV
import qualified Data.Vector as V
import qualified Data.List as List
import Data.Binary
import Data.Binary.Get
import Data.Binary.IEEE754
import Control.Monad
import qualified Data.Vector.Storable as SV
import MHUtils

foreign import ccall "gmmLProb" c_gmmLProb :: 
  CLong -> -- num mix
  Ptr CDouble ->  -- weights
  Ptr CDouble ->  -- concatenated mean vectors
  Ptr CDouble ->  -- concatenated var vectors
  Ptr CDouble ->  -- gconsts
  CLong -> -- feature dim
  Ptr CDouble ->  -- feature vector
  Double

gmmLProb :: GMixture -> FeaVector -> Double
gmmLProb gm fea
  = unsafePerformIO $
    SV.unsafeWith (weights gm) $ \wPtr ->
    SV.unsafeWith (means gm) $ \mPtr ->
    SV.unsafeWith (vars gm) $ \vPtr ->
    SV.unsafeWith (gconsts gm) $ \cPtr -> 
    SV.unsafeWith fea $ \feaPtr -> do
      return $ c_gmmLProb (fromIntegral (nmix gm)) wPtr mPtr vPtr cPtr (fromIntegral (SV.length fea)) feaPtr

-- HMM state observation density ------------------------
type FeaVector = SV.Vector CDouble
type GMixtures = V.Vector GMixture

data HtkFea = HtkFea {
  nSamples :: Word32,
  sampPeriod :: Word32,
  sampSize :: Word16,
  parmKind :: Word16,
  feaVec :: [FeaVector]
  }

data GMixture = GMixture {
  name :: [Char], 
  nmix :: Int, 
  weights :: SV.Vector CDouble,
  means :: SV.Vector CDouble,
  vars :: SV.Vector CDouble,
  gconsts :: SV.Vector CDouble
  } 

inspectSPDFs :: V.Vector GMixture -> Double
inspectSPDFs spdfs = V.foldl' (\acc gm -> acc + realToFrac (SV.foldl' (+) 0.0 (means gm))) 0.0 spdfs

spdfsFromFile :: String -> IO (V.Vector GMixture)
spdfsFromFile fname = do {buf <- L8.readFile fname; return (readSpdfs buf)}
        
readSpdfs :: L8.ByteString -> (V.Vector GMixture)
readSpdfs buf = V.fromList $ gmReadAll $ L8.words buf
  where gmReadAll buf | buf == [] = []
                      | otherwise = gmm:(gmReadAll rst)
          where (gmm, rst) = gmRead buf
        gmRead :: [L8.ByteString] -> (GMixture, [L8.ByteString])
        gmRead x | [x !! 0, x !! 2] == [L8.pack "~s", L8.pack "<NUMMIXES>"] = 
                                                     ((GMixture name
                                                     (mBtoI nmix)
                                                     (SV.fromList (List.map tplw mixTmp))
                                                     (SV.fromList $ List.concat $ List.map tplm mixTmp)
                                                     (SV.fromList $ List.concat $ List.map tplv mixTmp)
                                                     (SV.fromList (List.map tplc mixTmp))),
                                                    gmRst)
          where (mixTmp, gmRst) = gcNRead (mBtoI nmix) [] rst
                tplw (w,m,v,c) = w
                tplm (w,m,v,c) = m
                tplv (w,m,v,c) = v
                tplc (w,m,v,c) = c
		name = L8.unpack $ x !! 1
		nmix = x !! 3
		rst = drop 4 x
        gcNRead iter out buf | iter == 0 = (out, buf)
                             | otherwise = gcNRead (iter-1) ((mweight, mean, var, gconst):out) gcRst
          where (mweight, mean, var, gconst, gcRst) = gcRead buf
        gcRead x | x !! 0 == L8.pack "<MIXTURE>" = ((mBtoD mweight), mean, var, gconst, gconstRst)
          where (mean, meanRst) = meanRead rst
                (var, varRst) = varRead meanRst
                (gconst, gconstRst) = gconstRead varRst
		mid = x !! 1
		mweight = x !! 2
		rst = drop 3 x
        meanRead x | x !! 0 == L8.pack "<MEAN>" =  (mean, meanRst)
          where (mean, meanRst) = vecRead (mBtoI dim) rst
		dim = x !! 1
		rst = drop 2 x
        varRead x | x !! 0 == L8.pack "<VARIANCE>" = (var, varRst)
          where (var, varRst) = vecRead (mBtoI dim) rst
		dim = x !! 1
		rst = drop 2 x
        gconstRead x | x !! 0 == L8.pack "<GCONST>" = ((mBtoD gconst), rst)
	  where gconst = x !! 1
		rst = drop 2 x
        vecRead !dim buf = (List.map mBtoD (take dim buf), (drop dim buf))
          where (v, rst) = splitAt dim buf


instance Show HtkFea where
  show (HtkFea ns sp ss pk fv) = "HtkFea nSamples=" List.++ (show ns) List.++ " sampPeriod=" List.++ (show sp) 
                                 List.++ " sampSize=" List.++ (show ss) List.++ " parmKind=" List.++ (show pk)
                                 List.++ (List.foldl (\x y -> (x List.++ "\n" List.++ y)) "" (map show fv))

-- | Reads HTK format feature file
htkParmFromFile :: String -> IO HtkFea
htkParmFromFile fname = do {buf <- L8.readFile fname; return (runGet readHtkParm buf)}

readHtkParm :: Get (HtkFea)
readHtkParm = do {
  nSamples <- getWord32be;
  sampPeriod <- getWord32be;
  sampSize <- getWord16be;
  parmKind <- getWord16be;
  feas <- replicateM (fromIntegral nSamples) (readVec (fromIntegral (div sampSize 4)));
  return (HtkFea nSamples sampPeriod sampSize parmKind feas)
  }
  where readVec :: Int -> Get (SV.Vector CDouble)
        readVec dim = do {
          tmp <- replicateM dim getFloat32be;
          return (SV.fromList (List.map (fromRational.toRational) tmp))}

-- | Reads features in text format
txtFeaFromFile :: String -> IO [FeaVector]
txtFeaFromFile fname = do {buf <- L8.readFile fname; return (readTxtFea buf)}
                                                  
readTxtFea :: L8.ByteString -> [FeaVector]
readTxtFea buf = map (SV.fromList . (map mBtoD) . L8.words) $ (List.filter (\x -> x /= L8.empty)) $ L8.splitWith (=='\n') buf

